@@ -12,6 +12,7 @@
 #include "interval.h"
 #include "rule_p.h"
 #include "logging.h"
+#include "consecutiveaccumulator_p.h"
 
 #include <QDateTime>
 #include <QJsonArray>
@@ -52,48 +53,130 @@ void OpeningHoursPrivate::autocorrect()
         auto rule = (*it).get();
         auto prevRule = (*(std::prev(it))).get();
 
-        if (rule->m_ruleType != Rule::AdditionalRule || rule->hasComment() || prevRule->hasComment() || !prevRule->hasImplicitState()) {
+        if (rule->hasComment() || prevRule->hasComment() || !prevRule->hasImplicitState()) {
             continue;
         }
-
         const auto prevRuleSingleSelector = prevRule->selectorCount() == 1;
         const auto curRuleSingleSelector = rule->selectorCount() == 1;
 
-        // the previous rule has no time selector, the current rule only has a weekday selector
-        // so we fold the two rules together
-        if (!prevRule->m_timeSelector && prevRule->m_weekdaySelector && rule->m_weekdaySelector && !rule->hasWideRangeSelector()) {
-            auto tmp = std::move(rule->m_weekdaySelector);
-            rule->m_weekdaySelector = std::move(prevRule->m_weekdaySelector);
-            rule->m_weekSelector = std::move(prevRule->m_weekSelector);
-            rule->m_monthdaySelector = std::move(prevRule->m_monthdaySelector);
-            rule->m_yearSelector = std::move(prevRule->m_yearSelector);
-            rule->m_colonAfterWideRangeSelector = prevRule->m_colonAfterWideRangeSelector;
-            auto *selector = rule->m_weekdaySelector.get();
-            while (selector->rhsAndSelector)
-                selector = selector->rhsAndSelector.get();
-            appendSelector(selector, std::move(tmp));
-            rule->m_ruleType = prevRule->m_ruleType;
-            std::swap(*it, *std::prev(it));
-            it = std::prev(m_rules.erase(it));
+        if (rule->m_ruleType == Rule::AdditionalRule) {
+            // the previous rule has no time selector, the current rule only has a weekday selector
+            // so we fold the two rules together
+            if (!prevRule->m_timeSelector && prevRule->m_weekdaySelector && rule->m_weekdaySelector && !rule->hasWideRangeSelector()) {
+                auto tmp = std::move(rule->m_weekdaySelector);
+                rule->m_weekdaySelector = std::move(prevRule->m_weekdaySelector);
+                rule->m_weekSelector = std::move(prevRule->m_weekSelector);
+                rule->m_monthdaySelector = std::move(prevRule->m_monthdaySelector);
+                rule->m_yearSelector = std::move(prevRule->m_yearSelector);
+                rule->m_colonAfterWideRangeSelector = prevRule->m_colonAfterWideRangeSelector;
+                auto *selector = rule->m_weekdaySelector.get();
+                while (selector->rhsAndSelector)
+                    selector = selector->rhsAndSelector.get();
+                appendSelector(selector, std::move(tmp));
+                rule->m_ruleType = prevRule->m_ruleType;
+                std::swap(*it, *std::prev(it));
+                it = std::prev(m_rules.erase(it));
+            }
+
+            // the current rule only has a time selector, so we append that to the previous rule
+            else if (curRuleSingleSelector && rule->m_timeSelector && prevRule->m_timeSelector) {
+                appendSelector(prevRule->m_timeSelector.get(), std::move(rule->m_timeSelector));
+                it = std::prev(m_rules.erase(it));
+            }
+
+            // previous is a single monthday selector
+            else if (rule->m_monthdaySelector && prevRuleSingleSelector && prevRule->m_monthdaySelector && !isWiderThan(prevRule, rule)) {
+                auto tmp = std::move(rule->m_monthdaySelector);
+                rule->m_monthdaySelector = std::move(prevRule->m_monthdaySelector);
+                appendSelector(rule->m_monthdaySelector.get(), std::move(tmp));
+                rule->m_ruleType = prevRule->m_ruleType;
+                std::swap(*it, *std::prev(it));
+                it = std::prev(m_rules.erase(it));
+            }
+        } else if (rule->m_ruleType == Rule::NormalRule) {
+            // Previous rule has time and other selectors
+            // Current rule is only a time selector
+            // "Mo-Sa 12:00-15:00; 18:00-24:00" => "Mo-Sa 12:00-15:00,18:00-24:00"
+            if (curRuleSingleSelector && rule->m_timeSelector
+                    && prevRule->selectorCount() > 1 && prevRule->m_timeSelector) {
+                appendSelector(prevRule->m_timeSelector.get(), std::move(rule->m_timeSelector));
+                it = std::prev(m_rules.erase(it));
+            }
+
+            // Both rules have exactly the same selector apart from time
+            // Ex: "Mo-Sa 12:00-15:00; Mo-Sa 18:00-24:00" => "Mo-Sa 12:00-15:00,18:00-24:00"
+            // Obviously a bug, it was overwriting the 12:00-15:00 range.
+            // For now this only supports weekday selectors, could be extended
+            else if (rule->selectorCount() == prevRule->selectorCount()
+                     && rule->m_timeSelector && prevRule->m_timeSelector
+                     && !rule->hasComment() && !prevRule->hasComment()
+                     && rule->selectorCount() == 2 && rule->m_weekdaySelector && prevRule->m_weekdaySelector
+                     // slower than writing an operator==, but so much easier to write :)
+                     && rule->m_weekdaySelector->toExpression() == prevRule->m_weekdaySelector->toExpression()
+                     ) {
+                appendSelector(prevRule->m_timeSelector.get(), std::move(rule->m_timeSelector));
+                it = std::prev(m_rules.erase(it));
+            }
+        }
+    }
+}
+
+void OpeningHoursPrivate::simplify()
+{
+    if (m_error == OpeningHours::SyntaxError || m_rules.empty()) {
+        return;
+    }
+
+    for (auto it = std::next(m_rules.begin()); it != m_rules.end(); ++it) {
+        auto rule = (*it).get();
+        auto prevRule = (*(std::prev(it))).get();
+
+        if (rule->m_ruleType == Rule::AdditionalRule || rule->m_ruleType == Rule::NormalRule) {
+
+            auto hasNoHoliday = [](WeekdayRange *selector) {
+                return selector->holiday == WeekdayRange::NoHoliday
+                        && !selector->lhsAndSelector;
+            };
+            // Both rules have the same time and a different weekday selector
+            // Mo 08:00-13:00; Tu 08:00-13:00 => Mo,Tu 08:00-13:00
+            if (rule->selectorCount() == prevRule->selectorCount()
+                    && rule->m_timeSelector && prevRule->m_timeSelector
+                    && rule->selectorCount() == 2 && rule->m_weekdaySelector && prevRule->m_weekdaySelector
+                    && hasNoHoliday(rule->m_weekdaySelector.get())
+                    && hasNoHoliday(prevRule->m_weekdaySelector.get())
+                    && *rule->m_timeSelector == *prevRule->m_timeSelector
+                    ) {
+                // We could of course also turn Mo,Tu,We,Th into Mo-Th...
+                appendSelector(prevRule->m_weekdaySelector.get(), std::move(rule->m_weekdaySelector));
+                it = std::prev(m_rules.erase(it));
+                continue;
+            }
         }
 
-        // the current rule only has a time selector, so we append that to the previous rule
-        else if (curRuleSingleSelector && rule->m_timeSelector && prevRule->m_timeSelector) {
-            appendSelector(prevRule->m_timeSelector.get(), std::move(rule->m_timeSelector));
-            it = std::prev(m_rules.erase(it));
-        }
-
-        // previous is a single monthday selector
-        else if (rule->m_monthdaySelector && prevRuleSingleSelector && prevRule->m_monthdaySelector && !isWiderThan(prevRule, rule)) {
-            auto tmp = std::move(rule->m_monthdaySelector);
-            rule->m_monthdaySelector = std::move(prevRule->m_monthdaySelector);
-            appendSelector(rule->m_monthdaySelector.get(), std::move(tmp));
-            rule->m_ruleType = prevRule->m_ruleType;
-            std::swap(*it, *std::prev(it));
-            it = std::prev(m_rules.erase(it));
+        if (rule->m_ruleType == Rule::AdditionalRule) {
+            // Both rules have exactly the same selector apart from time
+            // Ex: "Mo 12:00-15:00, Mo 18:00-24:00" => "Mo 12:00-15:00,18:00-24:00"
+            // For now this only supports weekday selectors, could be extended
+            if (rule->selectorCount() == prevRule->selectorCount()
+                    && rule->m_timeSelector && prevRule->m_timeSelector
+                    && !rule->hasComment() && !prevRule->hasComment()
+                    && rule->selectorCount() == 2 && rule->m_weekdaySelector && prevRule->m_weekdaySelector
+                    // slower than writing an operator==, but so much easier to write :)
+                    && rule->m_weekdaySelector->toExpression() == prevRule->m_weekdaySelector->toExpression()
+                    ) {
+                appendSelector(prevRule->m_timeSelector.get(), std::move(rule->m_timeSelector));
+                it = std::prev(m_rules.erase(it));
+            }
         }
     }
 
+    // Now try collapsing adjacent week days: Mo,Tu,We => Mo-We
+    for (auto it = m_rules.begin(); it != m_rules.end(); ++it) {
+        auto rule = (*it).get();
+        if (rule->m_weekdaySelector) {
+            rule->m_weekdaySelector->simplify();
+        }
+    }
 }
 
 void OpeningHoursPrivate::validate()
@@ -326,6 +409,12 @@ QByteArray OpeningHours::normalizedExpression() const
     return ret;
 }
 
+QByteArray OpeningHours::simplifiedExpression() const
+{
+    d->simplify();
+    return normalizedExpression();
+}
+
 QString OpeningHours::normalizedExpressionString() const
 {
     return QString::fromUtf8(normalizedExpression());
